"""Annotate services

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
from os.path import basename
from typing import Tuple

from rougail.i18n import _
from rougail.utils import normalize_family
from rougail.error import DictConsistencyError
from rougail.annotator.variable import CONVERT_OPTION
try:
    import tiramisu3 as tiramisu
except ModuleNotFoundError:
    import tiramisu
# a object's attribute has some annotations
# that shall not be present in the exported (flatened) XML
ERASED_ATTRIBUTES = ('redefine', 'namespace', 'xmlfiles', 'disabled', 'name', 'manage')
ERASED_ATTRIBUTES2 = ('redefine', 'namespace', 'xmlfiles', 'disabled')
ALLOW_ATTRIBUT_NOT_MANAGE = ['file', 'engine', 'target', 'certificate']
FORCE_INFORMATIONS = ['mode']


class Annotator:
    """Manage service's object
    for example::
        <services>
            <service name="test">
                <service_access service='ntp'>
                </service_access>
            </service>
        </services>
    """
    level = 20
    def __init__(self,
                 objectspace,
                 *args,
                 ) -> None:
        self.objectspace = objectspace
        self.uniq_overrides = {}
        if 'network_type' not in self.objectspace.types:
            self.objectspace.types['network_type'] = self.objectspace.types['ip_type']
        services = []
        if not self.objectspace.paths.has_path_prefix():
            self.uniq_overrides[None] = []
            if hasattr(self.objectspace.space, 'services'):
                if not hasattr(self.objectspace.space.services, 'service'):
                    del self.objectspace.space.services
                else:
                    services.append((None, 'services', self.objectspace.space.services))
        elif hasattr(self.objectspace.space, 'variables'):
            for path_prefix in self.objectspace.paths.get_path_prefixes():
                self.uniq_overrides[path_prefix] = []
                root_path = f'{path_prefix}.services'
                if not path_prefix in self.objectspace.space.variables or \
                        not hasattr(self.objectspace.space.variables[path_prefix], 'services'):
                    continue
                if not hasattr(self.objectspace.space.variables[path_prefix].services, 'service'):
                    del self.objectspace.space.variables[path_prefix].services
                else:
                    services.append((path_prefix, root_path, self.objectspace.space.variables[path_prefix].services))
        for path_prefix, root_path, service in services:
            self.convert_services(path_prefix, root_path, service)

    def convert_services(self, path_prefix, root_path, services):
        """convert services to variables
        """
        services.hidden = True
        services.name = 'services'
        services.doc = 'services'
        services.path = root_path
        for service_name, service in services.service.items():
            service.name = normalize_family(service_name)
            activate_obj = self._generate_element('boolean',
                                                  None,
                                                  None,
                                                  'activate',
                                                  not service.disabled,
                                                  service,
                                                  f'{root_path}.{service.name}',
                                                  path_prefix,
                                                  )
            service.disabled = None
            dico = dict(vars(service))
            if 'type' not in dico:
                if service.manage:
                    dico['type'] = service.type
                else:
                    dico['type'] = 'none'
            for elttype, values in dico.items():
                if elttype == 'servicelist':
                    self.objectspace.paths.list_conditions[path_prefix].setdefault('servicelist',
                                                                                   {}).setdefault(
                                                                                       values,
                                                                                       []).append(activate_obj)
                    continue
                if elttype in ERASED_ATTRIBUTES:
                    continue
                if not service.manage and elttype not in ALLOW_ATTRIBUT_NOT_MANAGE and (elttype != 'type' or values != 'none'):
                    msg = _(f'unmanage service cannot have "{elttype}"')
                    raise DictConsistencyError(msg, 66, service.xmlfiles)
                if isinstance(values, (dict, list)):
                    if elttype != 'ip':
                        eltname = elttype + 's'
                    else:
                        eltname = elttype
                    if hasattr(service, 'servicelist'):
                        if isinstance(values, dict):
                            for key, value in values.items():
                                setattr(value, 'servicelist', service.servicelist)
                    family = self._gen_family(eltname,
                                              f'{root_path}.{service.name}',
                                              service.xmlfiles,
                                              path_prefix,
                                              with_informations=False,
                                              )
                    if isinstance(values, dict):
                        values = list(values.values())
                    family.family = self.make_group_from_elts(service_name,
                                                              elttype,
                                                              values,
                                                              f'{root_path}.{service.name}.{eltname}',
                                                              root_path,
                                                              path_prefix,
                                                              )
                    setattr(service, elttype, family)
                else:
                    if not hasattr(service, 'information'):
                        service.information = self.objectspace.information(service.xmlfiles)
                    setattr(service.information, elttype, values)
            service.path = f'{root_path}.{service.name}'
            manage = self._generate_element('boolean',
                                            None,
                                            None,
                                            'manage',
                                            service.manage,
                                            service,
                                            f'{root_path}.{service.name}',
                                            path_prefix,
                                            )
            service.variable = [activate_obj, manage]
            service.doc = service_name

    def make_group_from_elts(self,
                             service_name,
                             elttype,
                             elts,
                             path,
                             root_path,
                             path_prefix,
                             ):
        """Splits each objects into a group (and `OptionDescription`, in tiramisu terms)
        and build elements and its attributes (the `Options` in tiramisu terms)
        """
        families = []
        listname = '{}list'.format(elttype)
        for elt in elts:
            # try to launch _update_xxxx() function
            update_elt = '_update_' + elttype
            if hasattr(self, update_elt):
                getattr(self, update_elt)(elt,
                                          service_name,
                                          path_prefix,
                                          )
            c_name, subpath = self._get_name_path(elt,
                                                  path,
                                                  root_path,
                                                  path_prefix,
                                                  )
            family = self._gen_family(c_name,
                                      subpath.rsplit('.', 1)[0],
                                      elt.xmlfiles,
                                      path_prefix,
                                      )
            family.variable = []
            if hasattr(elt, 'disabled'):
                disabled = elt.disabled
            else:
                disabled = False
            activate_obj = self._generate_element('boolean',
                                                  None,
                                                  None,
                                                  'activate',
                                                  not disabled,
                                                  elt,
                                                  subpath,
                                                  path_prefix,
                                                  )
            for key in dir(elt):
                if key.startswith('_') or key.endswith('_type') or key in ERASED_ATTRIBUTES2:
                    continue
                value = getattr(elt, key)
                if key in [listname, 'servicelist']:
                    self.objectspace.paths.list_conditions[path_prefix].setdefault(key,
                                                                                   {}).setdefault(
                                                                                       value,
                                                                                       []).append(activate_obj)
                    continue
                if key == 'name':
                    dtd_key_type = elttype + '_type'
                else:
                    dtd_key_type = key + '_type'
                elt_type = getattr(elt, dtd_key_type, None)
                if elt_type:
                    try:
                        value = CONVERT_OPTION.get(elt_type, {}).get('func', str)(value)
                    except ValueError as err:
                        msg = _(f'"{value}" is not a valid "{elttype}": {err}')
                        raise DictConsistencyError(msg, 93, elt.xmlfiles)
                if key not in FORCE_INFORMATIONS and elt_type:
                    if elt_type == 'variable':
                        elt_type = 'symlink'
                    family.variable.append(self._generate_element(elt_type,
                                                                  dtd_key_type,
                                                                  elttype,
                                                                  key,
                                                                  value,
                                                                  elt,
                                                                  subpath,
                                                                  path_prefix,
                                                                  ))
                else:
                    setattr(family.information, key, value)

            family.variable.append(activate_obj)
            families.append(family)
        return families

    def _get_name_path(self,
                       elt,
                       path: str,
                       root_path: str,
                       path_prefix: str,
                       ) -> Tuple[str, str]:
        # create element name, if already exists, add _xx to be uniq
        if hasattr(elt, 'source') and elt.source:
            name = elt.source
        else:
            name = elt.name
        idx = 0
        while True:
            c_name = name
            if idx:
                c_name += f'_{idx}'
            subpath = '{}.{}'.format(path, normalize_family(c_name))
            try:
                self.objectspace.paths.get_family(subpath, root_path, path_prefix)
            except DictConsistencyError as err:
                if err.errno == 42:
                    return c_name, subpath
            idx += 1

    def _gen_family(self,
                    name,
                    subpath,
                    xmlfiles,
                    path_prefix,
                    with_informations=True,
                    ):
        family = self.objectspace.family(xmlfiles)
        family.name = normalize_family(name)
        family.doc = name
        family.mode = None
        self.objectspace.paths.add_family('services',
                                          subpath,
                                          family,
                                          False,
                                          force_path_prefix=path_prefix,
                                          )
        if with_informations:
            family.information = self.objectspace.information(xmlfiles)
        return family

    def _generate_element(self,
                          type_,
                          dtd_key_type,
                          elttype,
                          key,
                          value,
                          elt,
                          path,
                          path_prefix,
                          ):  # pylint: disable=R0913
        variable = self.objectspace.variable(elt.xmlfiles)
        variable.name = normalize_family(key)
        variable.mode = None
        variable.type = type_
        if type_ == 'symlink':
            variable.opt = self.objectspace.paths.get_variable(value,
                                                               self.objectspace.rougailconfig['variable_namespace'],
                                                               xmlfiles=elt.xmlfiles,
                                                               force_path_prefix=path_prefix,
                                                               add_path_prefix=True,
                                                               )
            variable.multi = None
            needed_type = self.objectspace.types[dtd_key_type]
            if elttype != 'certificate' and needed_type not in ('variable', variable.opt.type):
                msg = _(f'"{value}" in "{elttype}" must be a variable with type '
                        f'"{needed_type}" not "{variable.opt.type}"')
                raise DictConsistencyError(msg, 58, elt.xmlfiles)

        else:
            variable.doc = key
            variable.default = value
            variable.namespace = 'services'
        self.objectspace.paths.add_variable('services',
                                            path,
                                            variable,
                                            force_path_prefix=path_prefix
                                            )
        return variable

    def _update_override(self,
                         override,
                         service_name,
                         path_prefix,
                         ):

        if service_name in self.uniq_overrides[path_prefix]:
            msg = _('only one override is allowed by service, '
                    'please use a variable multiple if you want have more than one IP')
            raise DictConsistencyError(msg, 69, override.xmlfiles)
        self.uniq_overrides[path_prefix].append(service_name)
        override.name = service_name
        if not hasattr(override, 'source'):
            override.source = service_name

    @staticmethod
    def _update_file(file_,
                     service_name,
                     path_prefix,
                     ):
        if not hasattr(file_, 'file_type') or file_.file_type == "filename":
            if not hasattr(file_, 'source'):
                file_.source = basename(file_.name)
        elif not hasattr(file_, 'source'):
            msg = _(f'attribute "source" is mandatory for the file "{file_.name}" '
                    f'"({service_name})"')
            raise DictConsistencyError(msg, 34, file_.xmlfiles)

    def _update_ip(self,
                   ip,
                   service_name,
                   path_prefix,
                   ) -> None:
        variable = self.objectspace.paths.get_variable(ip.name,
                                                       ip.namespace,
                                                       xmlfiles=ip.xmlfiles,
                                                       force_path_prefix=path_prefix,
                                                       add_path_prefix=True,
                                                       )
        if variable.type not in ['ip', 'network', 'network_cidr']:
            msg = _(f'ip cannot be linked to "{variable.type}" variable "{ip.name}"')
            raise DictConsistencyError(msg, 70, ip.xmlfiles)
        if variable.type in ['ip', 'network_cidr'] and hasattr(ip, 'netmask'):
            msg = _(f'ip with ip_type "{variable.type}" must not have netmask')
            raise DictConsistencyError(msg, 59, ip.xmlfiles)
        if variable.type == 'network' and not hasattr(ip, 'netmask'):
            msg = _(f'ip with ip_type "{variable.type}" must have netmask')
            raise DictConsistencyError(msg, 64, ip.xmlfiles)
        if hasattr(ip, 'netmask'):
            netmask = self.objectspace.paths.get_variable(ip.netmask, 
                                                          ip.namespace,
                                                          xmlfiles=ip.xmlfiles,
                                                          force_path_prefix=path_prefix,
                                                          add_path_prefix=True,
                                                          )
            if netmask.type != 'netmask':
                msg = _(f'netmask in ip must have type "netmask", not "{netmask.type}"')
                raise DictConsistencyError(msg, 65, ip.xmlfiles)

    def _update_certificate(self,
                            certificate,
                            certificate_name,
                            path_prefix,
                            ) -> None:
        if certificate.certificate_type == "variable":
            variable = self.objectspace.paths.get_variable(certificate.name,
                                                           certificate.namespace,
                                                           xmlfiles=certificate.xmlfiles,
                                                           force_path_prefix=path_prefix,
                                                           add_path_prefix=True,
                                                           )
            certificate.catype = certificate.type
        if not hasattr(certificate, 'domain'):
            certificate.domain = self.objectspace.rougailconfig['default_cerficate_domain']
        variable = self.objectspace.paths.get_variable(certificate.domain,
                                                       certificate.namespace,
                                                       xmlfiles=certificate.xmlfiles,
                                                       force_path_prefix=path_prefix,
                                                       add_path_prefix=True,
                                                       )
        if variable.type != 'domainname':
            msg = _(f'the certificate "{certificate.name}" has an attribute "domain" linked with a "{variable.type}" variable ("{certificate.domain}"), but must be a "domainename" variable')
            raise DictConsistencyError(msg, 94, certificate.xmlfiles)
